//---------------------------------------------------------------------------
//
// Evj        Model the evolution of DNA binding site recognition
//
// Module:    Param, the simulation parameter class
// Author:    Paul C. Anagnostopoulos
// Created:   1 April 2005
//
// This computer program fits the description in the U.S. Copyright Act
// of a "United States Government work." It was written as a part of
// my/our official duties as a Government employee. This means it cannot
// be copyrighted. The program is freely available for use without a
// copyright notice, and there are no restrictions on its use, now or
// subsequently. We retain no rights in the program. Please acknowledge
// the source of this code in any releases you make.
//
//---------------------------------------------------------------------------


import java.awt.*;
import java.util.*;
import java.util.prefs.*;
import javax.swing.*;

//                      Parameter Definitions
//                      --------- -----------


public final class Param
{
  public static final int
    version               =  0,
    title                 =  1,
    population            =  2,
    potentialSites        =  3,
    sites                 =  4,
    siteWidth             =  5,
    placement             =  6,
    weightWidth           =  7,
    mutationMethod        =  8,
    mutationsPerGenome    =  9,
    mutationsPerBases     = 10,
    selection             = 11,
    tie                   = 12,
    cycles                = 13,
    seed                  = 14,
    perfection            = 15,
    convergence           = 16,
    perfecgence           = 17,
                                  // Advanced parameters:
    useFounder            = 18,
    founders              = 19,
    missedSite            = 20,
    spuriousGene          = 21,
    spuriousNongene       = 22;
  
  public static final char PLACE_ARRAYED        = 'A',
                           PLACE_RANDOM         = 'N',
                           PLACE_RANDOM_OVERLAP = 'O';

  public static final char MUTATION_PER_GENOME = 'G',
                           MUTATION_PER_BASES  = 'B';
  
  public static final char TIE_BOTH   = 'B',
                           TIE_FIRST  = 'F',
                           TIE_RANDOM = 'R';
  
  // Set up an array of parameter definitions. Each element of the
  // array is a ParamDef object.

  private static final ParamDef paramInfo[] = {
    new ParamDef("version",          "Evj version",
                                     'S', ""),
    new ParamDef("title",            "title of model",
                                     'S', ""),
    new ParamDef("population",       "number of creatures",
                                     'I', 64, 2, Integer.MAX_VALUE, 2),
    new ParamDef("genome-size",      "bases per genome, G",
                                     'I', 256, 48, Integer.MAX_VALUE, 16),
    new ParamDef("sites",            "binding sites per genome, gamma",
                                     'I', 16, 1, Integer.MAX_VALUE, 1),
    new ParamDef("site-width",       "bases per binding site, L",
                                     'I', 6, 1, 100, 1),
    new ParamDef("placement",        "binding site placement method",
                                     'C', PLACE_ARRAYED, "ANO"),
    new ParamDef("weight-width",     "bases per weight in matrix",
                                     'I', 5, 1, 15, 1),
    new ParamDef("mutation-method",  "mutation method",
                                     'C', MUTATION_PER_GENOME, "BG"),
    new ParamDef("mutations-per-genome", "mutations per genome",
                                     'I', 1, 0, Integer.MAX_VALUE, 1),
    new ParamDef("mutations-per-bases", "1 mutation per bases",
                                     'I', 256, 1, Integer.MAX_VALUE, 8),
    new ParamDef("selection",        "perform selection",
                                     'B', true),
    new ParamDef("tie",              "special rule for tie scores",
                                     'C', TIE_BOTH, "BFR"),
    new ParamDef("cycles",           "number of cycles to model",
                                     'I', 10000, 1, 99000000, 10000),
    new ParamDef("seed",             "random number generator seed",
                                     'I', 0, -1000000, 1000000, 1),
    new ParamDef("perfection",       "pause on 0 mistakes",
                                     'B', false),
    new ParamDef("convergence",      "pause on Rfreq >= Rseq",
                                     'B', false),
    new ParamDef("perfecgence",      "pause on both conditions",
                                     'B', false),
    new ParamDef("founder-model",    "use founder model",
                                     'B', false),
    new ParamDef("founders",         "number of founder creatures",
                                     'I', 1, 1, Integer.MAX_VALUE, 1),
    new ParamDef("missed-site",      "missed site points",
                                     'I', 1, 0, 100, 1),
    new ParamDef("spurious-gene",    "spurious gene points",
                                     'I', 1, 0, 100, 1),
    new ParamDef("spurious-nongene", "spurious nongene points",
                                     'I', 1, 0, 100, 1)
    };

  private Object param[];           // Value of each parameters.

  private Object savedParam[];      // Copy of initial values in case
                                    // they are modified.

//                      Constructors
//                      ------------


  // This constructor returns a Param with the default values for every
  // parameter. It also saves the values so they can be restored.
  
  public Param ()
  {
    param = new Object[paramInfo.length];
    for (int i = 0; i < paramInfo.length; ++i) param[i] = paramInfo[i].def;
    param[version] = Evj.facilityVersion;
    save_for_restore();
  }

  // This method copies the values from one parameter object to
  // another, also saving them for restore.

  public void copy (Param source)
  {
    param = (Object[])(source.param.clone());
    save_for_restore();
  }

  // This method saves the values in a second array so they can be
  // restored later if necessary.

  public void save_for_restore ()
  {
    savedParam = (Object[])(param.clone());
  }
  
  // This method restores the saved parameter values.

  public void reset ()
  {
    param = (Object[])(savedParam.clone());
  }

//                      Validate Parameters
//                      -------- ----------


  // This method validates the parameters after they have been
  // populated. It returns an error message if there is a problem,
  // null otherwise.

  public String validate ()
  {
    int pop    = ((Integer)param[population]).intValue(),
        site   = ((Integer)param[sites]).intValue(),
        swidth = ((Integer)param[siteWidth]).intValue(),
        wwidth = ((Integer)param[weightWidth]).intValue(),
        size   = ((Integer)param[potentialSites]).intValue() + swidth - 1,
        found  = ((Integer)param[founders]).intValue(),
        siteSpace;
    
    // The population must be an even number.

    if (pop % 2 != 0)
      return Resource.get("pvmsg.evenPopulation");
      
    // If we're using a founder model, the number of founders must be
    // no more than half the population.

    if (getBoolean(useFounder) && found > pop/2)
      return Resource.get("pvmsg.founders1") + found +
             Resource.get("pvmsg.founders2") + " " + pop + " " +
             Resource.get("pvmsg.founders3");
    
    // The weighting matrix, threshold, and binding sites must fit in
    // the chromosome with a little space to spare. With overlapped
    // placement, we require enough space for a maximum 50% overlap.

    if (getCode(placement) != PLACE_RANDOM_OVERLAP)
      siteSpace = site * swidth;
    else
      siteSpace = site * swidth / 2;
    if (4 * wwidth * swidth +
        wwidth +
        siteSpace > .92 * size)
      return Resource.get("pvmsg.smallSize1") + " " + size + " " +
             Resource.get("pvmsg.smallSize2") + " " + site + " " + swidth +
             Resource.get("pvmsg.smallSize3");
    
    // The maximum weighting value must fit in a Java integer.

    if (2 * wwidth - 1 + Math.ceil(Stats.log2(swidth)) >
        Math.ceil(Stats.log2(Integer.MAX_VALUE)))
      return Resource.get("pvmsg.overflow1") + " " + wwidth + " " +
          Resource.get("pvmsg.overflow2") + " " + swidth + " " +
          Resource.get("pvmsg.overflow3");
    
    return null;
  }

//                      Get Parameters
//                      --- ----------


  // These public methods are used to obtain parameter values.
  
  public boolean getBoolean (int index)
  {
    return ((Boolean)param[index]).booleanValue();
  }

  public char getCode (int index)
  {
    return ((Character)param[index]).charValue();
  }

  public int getInt (int index)
  {
    return ((Integer)param[index]).intValue();
  }

  public String getString (int index)
  {
    return (String)param[index];
  }

  public boolean isDefaultAdvanced ()
  {
    return param[useFounder].equals(paramInfo[useFounder].def) &&
           param[missedSite].equals(paramInfo[missedSite].def) &&
           param[spuriousGene].equals(paramInfo[spuriousGene].def) &&
           param[spuriousNongene].equals(paramInfo[spuriousNongene].def);
  }

//                      Set Parameters
//                      --- ----------


  // These public methods are used to set parameter values.

  public void setBoolean (int index, boolean value)
  {
    param[index] = new Boolean(value);
  }

  public void setInteger (int index, int value)
  {
    param[index] = new Integer(value);
  }

//                      Populate GUI Components
//                      -------- --- ----------


  public void populateButtonGroup (int index, ButtonGroup group,
                                   boolean set)
  {
    if (set) {
      char code = ((Character)param[index]).charValue();
      for (Enumeration e = group.getElements(); e.hasMoreElements(); ) {
        ButtonModel model = ((JRadioButton)e.nextElement()).getModel();
        model.setSelected(model.getActionCommand().charAt(0) == code);
      }

    } else {
      ButtonModel model = group.getSelection();
      param[index] = new Character(model.getActionCommand().charAt(0));
    }
  }

  public void populateCheckBox (int index, JCheckBox box, boolean set)
  {
    if (set) {
      box.setSelected(((Boolean)param[index]).booleanValue());

    } else {
      param[index] = new Boolean(box.isSelected());
    }
  }
  
  public void populateIntegerField (int index, JTextField field, boolean set)
  {
    if (set) {
      field.setText(String.valueOf(param[index]));

    } else {
      param[index] = Integer.valueOf(field.getText());
    }
  }

  public void populateSpinner (int index, JSpinner spinner, boolean set)  
  {
    SpinnerNumberModel model = (SpinnerNumberModel)(spinner.getModel());
    
    if (set) {
      model.setMinimum(new Integer(paramInfo[index].min));
      model.setMaximum(new Integer(paramInfo[index].max));
      model.setStepSize(new Integer(paramInfo[index].step));
      model.setValue(param[index]);
      
    } else {
      param[index] = model.getValue();
    }
  }

  public void populateTextField (int index, JTextField field, boolean set)
  {
    if (set) {
      field.setText((String)(param[index]));

    } else {
      param[index] = field.getText();
    }
  }

//                      Derived Parameters
//                      ------- ----------


  // These statistics are derived from the specified parameters. They
  // are calculated in the Stats class.

  // The number of bits of information needed to bind the sites.

  public double Rfreq;

  // The expected uncertainty of the number of sample bases [E(Hnb)],
  // in bits/base.

  public double E_Hnb;

//                      Save Parameters
//                      ---- ----------


  // This method saves parameters in a preferences block whose name is
  // specified. Each parameter is saved according to its type.
  
  public void saveParameters (String name)
  {
    Preferences pref =
      Preferences.userRoot().node(Evj.preferencesRoot + name);

    for (int i = 0; i < paramInfo.length; ++i) {
      switch (paramInfo[i].type) {
      case 'B': pref.putBoolean(paramInfo[i].key,
                                ((Boolean)param[i]).booleanValue()); break;
      case 'C': pref.put(paramInfo[i].key,
                         ((Character)param[i]).toString()); break;
      case 'I': pref.putInt(paramInfo[i].key,
                            ((Integer)param[i]).intValue()); break;
      case 'S': pref.put(paramInfo[i].key,
                         (String)param[i]); break;
      }
    }

///    pref.flush();
  }

//                      Load Parameters
//                      ---- ----------


  // This method loads a Param object with the parameters saved in a
  // specified preferences block. It returns true if the loading was
  // successful, false otherwise.
  
  public boolean loadParameters (String name)
  {
    Preferences root = Preferences.userRoot(),
                pref;

    try {
      if (!root.nodeExists(Evj.preferencesRoot + name)) return false;
    } catch (Exception ex) {
    }

    pref = Preferences.userRoot().node(Evj.preferencesRoot + name);

    for (int i = 0; i < paramInfo.length; ++i) {
      switch (paramInfo[i].type) {
        case 'B': param[i] = new Boolean(pref.getBoolean(paramInfo[i].key,
                                                         false)); break;
        case 'C': param[i] = new Character(pref.get(paramInfo[i].key,
                                                    "").charAt(0)); break;
        case 'I': param[i] = new Integer(pref.getInt(paramInfo[i].key,
                                                     0)); break;
        case 'S': param[i] = pref.get(paramInfo[i].key, ""); break;
      }
    }

    save_for_restore();
    return true;
  }
}

//                      Definition Record
//                      ---------- ------


// This helper class defines a record to store the definition of one
// simulation parameter. 

final class ParamDef
{
  String key;                   // Parameter keyword.
  String description;           // Description of parameter.
  char type;                    // Type.
  Object def;                   // Default value.
  int min, max;                 // Minimum and maximum for integers.
  int step;                     // Increment for integers.
  String codes;                 // List of valid codes.

  ParamDef (String key, String description, char type, Object def)
  {
    this.key         = key;
    this.description = description;
    this.type        = type;
    this.def         = def;
  }

  ParamDef (String key, String description, char type, boolean def)
  {
    this(key, description, type, new Boolean(def));
  }

  ParamDef (String key, String description, char type, int def,
            int min, int max, int step)
  {
    this(key, description, type, new Integer(def));
    this.min  = min;
    this.max  = max;
    this.step = step;
  }

  ParamDef (String key, String description, char type, char def,
            String codes)
  {
    this(key, description, type, new Character(def));
    this.codes       = codes;
  }
}
